"""External Storage for Archetypes.

Implements a filesystem based storage for archetypes fields.

Variable used in instance:

    _v_olddata -- temporary data when copying/cutting/pasting
    _es_archive -- holds the archive history/versions
    _es_initialized -- defines if the external storage is initialized (bool)
    _es_filename -- filename of the external file
    _es_filepath -- path to the file relative to base directory
    _es_mimetype -- content type of the external file

    (I wonder if some of this can also be directly stored inside the storage object which
     is attached to the actual instance by Object -> Schema -> Field -> Storage)

It only supports one field defining an external storage per class.

"""

import os

from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from Globals import INSTANCE_HOME
from DateTime import DateTime
from StringIO import StringIO
from AccessControl import ClassSecurityInfo

from OFS.Image import File as OFSFile
from OFS.content_types import guess_content_type
from ZODB.PersistentMapping import PersistentMapping

from Products.CMFCore.utils import getToolByName
from Products.Archetypes.Storage import StorageLayer

from Products.Archetypes.interfaces.layer import ILayer
from Products.Archetypes.interfaces.field import IObjectField
from Products.ExternalStorage.interfaces import IExternalStorage

from Products.ExternalStorage.utils import log
from Products.ExternalStorage.utils import ERROR
from Products.ExternalStorage.utils import cleanupFilename

from Products.ExternalStorage.FileUtils import copy_file

from Products.ExternalStorage.config import ARCHIVE_PREFIX
from Products.ExternalStorage.config import ENVIRONMENT_VARIABLE

try:
    from ZPublisher.Iterators import filestream_iterator
    HAS_ITERATOR = 1
except ImportError:
    HAS_ITERATOR = 0

try:
    True
except:
    True = 1
    False = 0


class VersionInfo:
    """Models a version information for the file archive.
    """

    def __init__(self, vid, path, length, id, date=None):
        """Initializes a version information.

        Input parameters:

            vid    - version id
            date   - actual date in DateTime
            path   - path to that file relative to var/
            length - length of the file
            id     - id of the file
        """
        if date is None:
            date = DateTime()
        self.vid = vid
        self.date = date
        self.path = path
        self.length = length
        self.id = id


class ExternalStorage(StorageLayer):
    """Stores data as an external file.

    It gets stored in var/<configureddir>/<path>
    For now they all go into one file repository without directories.
    Archived versions will be stored in var/<configureddir>/ARCHIVE_PREFIX/<path>
    """

    __implements__ = (IExternalStorage, ILayer)

    security = ClassSecurityInfo()

    def __init__(self, prefix='files', archive=True, rename=False):
        """Initializes default values.
        """
        self.prefix = prefix
        self.archive = archive
        self.rename = rename
        self.renaming = False
        self._v_olddata = {}

    #
    # Storage API
    #

    security.declarePrivate('initializeInstance')
    def initializeInstance(self, instance, item=None, container=None):
        """Initializes a file.
        """
        log("initializeInstance")
        if hasattr(aq_base(instance), '_es_initialized'):
            # Already initialized, but it could be that the filepath
            # has changed, which might happen on a copy
            bp = self.getBasePath()
            p1 = bp + '/' + self.computeFilePath(instance)
            p2 = self.getFilepath(instance)
            if p1 == p2:
                log("already initialized %s" % p1)
                # no change, finished
                return
            else:
                # Store data in _v_olddata and let the rest be done by
                # the normal copy routine
                log("assuming a copy as the filepath has changed from %s to %s" % (p1, p2))
                name = self.getName()
                self._v_olddata[instance.UID()] = self.get(name, instance)

        instance._es_initialized = True

        # Set (new) filepath
        # XXX: compute this differently and make archive like this as well
        # XXX: Archivepath: external/ARCHIVE_PREFIX/<path>/
        # problem: when moving files around the archive is not found anymore..
        # but maybe we can store the path in the archivefield in the type
        p = self.computeFilePath(instance)
        self.setFilepath(p, instance)

        # Eventually create directory for file
        datadir = os.path.dirname(self.getFilepath(instance))
        if not os.path.exists(datadir):
            log("creating directory: %s" % datadir)
            os.makedirs(datadir)
        else:
            log("directory already exists: %s" % datadir)

        # Create archive, if necessary
        if self.archive:
            if not hasattr(aq_base(instance), '_es_archive'):
                log("initializing archive")
                instance._es_archive = PersistentMapping()

            # Eventually create directory for archive
            bp = self.getBasePath()
            fp = self.computeFilePath(instance, doArchive=1)
            arcdatadir = os.path.join(bp, os.path.dirname(fp))
            if not os.path.exists(arcdatadir):
                log("creating directory for %s" % arcdatadir)
                os.makedirs(arcdatadir)

        # Check for temp var
        uid = instance.UID()
        if self._v_olddata.has_key(uid):
            log("temp var found, setting them")
            # There is an old version there (copy/cut/paste)
            olddata = self._v_olddata[uid]
            # Make sure to not fire rename process again after add the object
            self.renaming = True
            self.set(self.getName(), instance, olddata)
            self.renaming = False
            del self._v_olddata[uid]
        else:
            log("no temp var found, setting initial version")
            # Initial version number
            self.setVersion(0, instance)

    security.declarePrivate('cleanupInstance')
    def cleanupInstance(self, instance, item=None, container=None):
        """Cleanup all stuff for that file.
        """
        log("cleanupInstance %s" % instance)
        # Set to uninitialized state
        if hasattr(aq_base(instance), '_es_initialized'):
            delattr(instance, '_es_initialized')
        # XXX: archive?!?

    security.declarePrivate('initializeField')
    def initializeField(self, instance, field):
        log("initializeField: %s" % field)

    security.declarePrivate('cleanupField')
    def cleanupField(self, instance, field, **kwargs):
        """Cleanup field's content on filesystem.
        """
        log("cleanupField: %s.%s" % (instance, field))
        self.unset(field.getName(), instance, **kwargs)
        filepath = self.getFilepath(instance)
        if filepath is not None:
            if os.path.exists(filepath):
                os.remove(filepath)
        else:
            # XXX: Do some better error handling!
            print "*** filepath not existant at unset?!?"

    security.declarePrivate('get')
    def get(self, name, instance, **kwargs):
        """Returns the contents of a file.
        """
        log("get: %s" % name)
        path = self.getFilepath(instance)
        if not os.path.exists(path):
            return ""
        if kwargs.get('useIterator', None) is not None and HAS_ITERATOR:
            return filestream_iterator(path, 'rb')
        else:
            data = StringIO()
            copy_file(path, data)
            data.seek(0)
            filename = self.getFilename(instance)
            mimetype = self.getContentType(instance)
            if filename and mimetype:
                data = OFSFile(id=name, title='', file=data, content_type=mimetype)

                data.filename = filename
                return data
            return data.getvalue()

    security.declarePrivate('set')
    def set(self, name, instance, value, **kwargs):
        """Stores some data as external file using id.
        """
        filename = getattr(value, 'filename', '')
        log("set %s: %s" % (name, filename))

        # Remove the 'illegal' characters from the filename
        # to avoid a Zope Copy Error Exception
        filename = cleanupFilename(filename)

        if filename:
            filename = filename.split('\\')[-1]
            # Check if the operation should trigger the object rename process
            if self.rename and not self.renaming and filename != instance.getId():
                # Call plone rename
                plone_utils = getToolByName(instance, 'plone_utils')
                plone_utils.contentEdit(instance, id=filename)

        if IObjectField.isImplementedBy(value):
            new_value = value.getRaw(instance)
        elif isinstance(value, OFSFile):
            new_value = value.data
        else:
            new_value = str(value)

        if new_value is None:
            new_value = ''

        data = StringIO(new_value)

        mimetype = kwargs.get('mimetype', 'application/octet-stream')

        instance._es_mimetype = mimetype
        instance._es_filename = filename

        try:
            copy_file(data, self.getFilepath(instance))
        except IOError, e:
            log("copy of external file failed", text=str(e), log_level=ERROR)
            raise

        # Copy to archive
        if self.archive:
            vid = self.incrementVersion(instance)
            afp = self.getArchiveFilepath(vid, instance)
            # XXX: Can we get an error here due to some file which might have the same name?
            # One should better check... but what then??
            # OTOH the UID is used as filename, thus everything should be ok!
            copy_file(data, afp)

            # Remember as (date, arcfilename, size, filename)
            instance._es_archive[vid] = ((DateTime(), afp, len(new_value), instance.getId()))

    security.declarePrivate('unset')
    def unset(self, name, instance, **kwargs):
        """Removes the field's content.
        """
        log("unset field %s of instance %s" % (name, instance))
        # Save old value
        v = self.get(name, instance)
        self._v_olddata[instance.UID()] = v

    #
    # Private Methods
    #

    security.declarePrivate('getBasePath')
    def getBasePath(self):
        """Returns the base path for this instance.

        This is usually the 'var' directory in the INSTANCE_HOME.

        You can also override this by setting a environment variable
        which is then used instead.
        """
        base = INSTANCE_HOME + '/var'
        if os.environ.has_key(ENVIRONMENT_VARIABLE):
            base = os.environ.get(ENVIRONMENT_VARIABLE)
        path = os.path.join(base, self.prefix)
        path = os.path.normpath(path)
        log('getBasePath: %s' % path)
        return path

    security.declarePrivate('computeFilePath')
    def computeFilePath(self, instance, doArchive=None):
        """Computes the current instance path (relative to portal).
        """
        prefix = ''
        if doArchive is not None:
            prefix = ARCHIVE_PREFIX
        pu = getToolByName(instance, 'portal_url')
        relative = pu.getRelativeContentURL(instance)
        path = os.path.join(prefix, relative)
        path = os.path.normpath(path)
        log("computeFilepath: %s" % path)
        return path

    security.declarePrivate('getFilepath')
    def getFilepath(self, instance):
        """Returns the path to the file.
        """
        bp = self.getBasePath()
        rp = getattr(aq_base(instance), '_es_filepath', '')
        path = os.path.normpath(rp)
        if path[0] != '/':
            path = os.path.join(bp, path)
        log("getFilepath: %s" % path)
        return path

    security.declarePrivate('getArchiveFilepath')
    def getArchiveFilepath(self, vid, instance):
        """Returns the path to the file in archive.
        """
        bp = self.getBasePath()
        rp = os.path.dirname(self.computeFilePath(instance, doArchive=1))
        path = os.path.normpath(rp)
        if path[0] != '/':
            path = os.path.join(bp, path)
        uid = instance.UID()
        afp = os.path.join(path, uid + ".backup." + vid)
        log("getArchiveFilepath: %s" % afp)
        return afp

    security.declarePrivate('setFilepath')
    def setFilepath(self, filepath, instance):
        """Sets the path to the file.
        """
        log("setFilepath: %s" % filepath)
        instance._es_filepath = filepath

    security.declarePrivate('setArchiveFilepath')
    def setArchiveFilepath(self, filepath, instance):
        """Sets the path to the file in archive.
        """
        log("setArchiveFilepath: %s" % filepath)
        instance._es_archive_filepath = filepath

    security.declarePrivate('getArchive')
    def getArchive(self, instance):
        """Returns the archive list.
        """
        archive = getattr(aq_base(instance), '_es_archive', None)
        log("getArchive: %s" % archive)
        return archive

    security.declarePrivate('getFilename')
    def getFilename(self, instance, name=None):
        """Returns the filename of the file.
        """
        default = None
        filename = getattr(aq_base(instance), '_es_filename', default)
        log("getFilename: %s" % filename)
        return filename

    security.declarePrivate('getContentType')
    def getContentType(self, instance, name=None):
        """Returns the content type of the file.
        """
        default = 'plain/text'
        mimetype = getattr(aq_base(instance), '_es_mimetype', default)
        log("getContentType: %s" % mimetype)
        return mimetype

    #
    # Version support
    #

    security.declarePrivate('setVersion')
    def setVersion(self, vid, instance):
        """Sets a version in the instance.
        """
        instance._es_version_id = vid

    security.declarePrivate('getVersion')
    def getVersion(self, instance, default=0):
        """Gets back the version number.
        """
        return getattr(aq_base(instance), '_es_version_id', default)

    security.declarePrivate('incrementVersion')
    def incrementVersion(self, instance):
        """Increments the version id.
        """
        vid = self.getVersion(instance)
        vid = vid + 1
        self.setVersion(vid, instance)
        return str(vid)

    security.declarePrivate('getArchiveVersion')
    def getArchiveVersion(self, vid, instance):
        """Copy an archived version back to the instance.
        """
        archive = self.getArchive(instance)
        if archive is None:
            return
        dt, vid, l, fn = archive[str(vid)]

        # Get data
        data = StringIO()
        path = self.getArchiveFilepath(vid, instance)
        copy_file(path, data)
        data = data.getvalue()

        # We don't copy it, as it has to be done via the upper layer!
        return data

    # XXX: Maybe an archived version can be retrieved without directly calling
    #      the storage maybe by applying some kwargs parameters (version=...)
    #      though we still need to get a list of versions etc.. so maybe some
    #      mixin class would do the trick...
